'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var types = {
    Object: function (val) { return !!val && !Array.isArray(val) && typeof val === "object"; },
    Number: function (val) { return typeof val === "number"; },
    String: function (val) { return typeof val === "string"; },
    Boolean: function (val) { return val === true || val === false; },
    Array: Array.isArray,
    Any: function (val) { return val !== undefined && val !== null; },
}

// TODO: OneOfType takes array instead of args?
// TODO: OneOf takes object and matches values
// TODO: types.Flag === types.Boolean.optional

var optional = function (fn) { return function (val) { return val === undefined || val === null || fn(val); }; }

for (var key in types) {
    types[key].optional = optional(types[key])
}

var tParams = {
    OneOf: function (opts) { return function (val) { return opts.indexOf(val) !== -1; }; },
    ArrayOf: function (typeFn) { return function (val) { return Array.isArray(val) && !!val.every(typeFn); }; },
    OneOfType: function () {
        var args = [], len = arguments.length;
        while ( len-- ) args[ len ] = arguments[ len ];

        return function (val) { return args.some(function (test) { return test(val); }); }
    },
}

var compose = function (a, b) { return function () {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    return a(b.apply(this, args))
}; }

for (var key$1 in tParams) {
    types[key$1] = tParams[key$1]
    types[key$1].optional = compose(optional, tParams[key$1])
}

var defaultMiddlewareOptions = {
    ignorePayloads: false,
    onError: console.error.bind(console, "unknown action:"),
    ignoreActions: ["EFFECT_TRIGGERED", "EFFECT_RESOLVED", "@@router/UPDATE_LOCATION"],
}

var merge$1 = function (a, b) { return Object.assign({}, a, b); }

// TODO: separate onError for unknown actions & bad props

var middlewareHelper = function (tests, unformat) { return function (options) {
    if ( options === void 0 ) options = {};

    var ref = merge$1(defaultMiddlewareOptions, options);
    var ignoreActions = ref.ignoreActions;
    var ignorePayloads = ref.ignorePayloads;
    var onError = ref.onError;
    var ignoreMap = ignoreActions.reduce(function (obj, key) { obj[key] = true; return obj }, {})

    var test = function (action) {
        var ref = unformat(action);
        var type = ref.type;
        var payload = ref.payload;
        if (ignoreMap[type]) { return }
        if (tests[type] && ignorePayloads) { return }
        if (tests[type] && tests[type](payload)) { return }
        onError(action)
    }

    return function () { return function (next) { return function (action) {
        test(action)
        return next(action)
    }; }; }
}; }

// TODO: throw custom functions

var reducerHelper = function (actions, unformat) { return function (obj, initState) {
    var nObj = {}
    for (var key in obj) {
        var nKey = actions[key]
        if (!nKey) { throw new Error(("unknown action: " + key)) }
        var fn = obj[key]
        if (typeof fn !== "function") { throw new Error((key + " is not a function")) }
        nObj[nKey] = obj[key]
    }

    return function (state, action) {
        if ( state === void 0 ) state = initState;

        var ref = unformat(action);
        var type = ref.type;
        var payload = ref.payload;
        return nObj[type]
            ? nObj[type](state, payload, action)
            : state
    }
}; }

var uncons = function (array) { return [array[0], array.slice(1)]; }

function parseAction (action) {
    var ref = uncons(action);
    var type = ref[0];
    var docAndArgs = ref[1];
    if (!docAndArgs.length) { return { type: type, args: [], doc: "" } }

    if (typeof docAndArgs[0] === "string") {
        var ref$1 = uncons(docAndArgs);
        var doc = ref$1[0];
        var args = ref$1[1];
        return { type: type, doc: doc, args: args.map(parseArg) }
    }
    return { type: type, doc: "", args: docAndArgs.map(parseArg) }
}

function parseArg (arg, i) {
    if (typeof arg === "function" && i === 0) {
        return { test: arg, doc: "", wholePayload: true }
    }
    if (arg.length === 3) {
        var id$1 = arg[0];
        var doc = arg[1];
        var test$1 = arg[2];
        return { id: id$1, doc: doc, test: test$1 }
    }
    var id = arg[0];
    var test = arg[1];
    return { id: id, doc: "", test: test }
}

var merge = function (a, b) { return Object.assign({}, a, b); }

var defaultParams = {
    format: function (type, payload) { return ({ type: type, payload: payload }); },
    unformat: function (action) { return action; },
    namespace: "",
}

// TODO: combineSchema
// TODO: namespace takes fn

function makeSchema (schema, params) {
    if ( params === void 0 ) params = {};

    var ref = merge(defaultParams, params);
    var format = ref.format;
    var unformat = ref.unformat;
    var namespace = ref.namespace;
    var parsed = schema.map(parseAction)

    var schemaMap = parsed.reduce(function (obj, action) {
        obj[action.type] = action
        return obj
    }, {})

    // action type -> namespaced action type
    var actions = parsed.reduce(function (obj, ref) {
        var type = ref.type;

        var nType = namespace
            ? namespace + "_" + type
            : type
        obj[type] = nType
        return obj
    }, {})

    // action type -> payload => namespaced action
    var actionCreators = parsed.reduce(function (obj, ref) {
        var type = ref.type;
        var args = ref.args;

        var nType = actions[type]
        var ac = function (payload) { return format(nType, payload); }
        ac.byPosition = function (a, b, c) {
            var arg0 = args[0];
            var arg1 = args[1];
            var arg2 = args[2];

            if (!args.length) { return format(nType) }
            if (arg0.wholePayload) { return format(nType, a) }

            var payload = {}
            payload[arg0.id] = a
            if (arg1) { payload[arg1.id] = b }
            if (arg2) { payload[arg2.id] = c }
            return format(nType, payload)
        }
        obj[type] = ac
        return obj
    }, {})

    // namespaced action type -> test
    var tests = parsed.reduce(function (obj, ref) {
        var type = ref.type;
        var args = ref.args;

        var nType = actions[type]
        if (!args.length) {
            obj[nType] = function (payload) { return payload === undefined; }
        } else if (args.length === 1 && args[0].wholePayload) {
            obj[nType] = args[0].test
        } else {
            obj[nType] = function (payload) { return payload &&
                typeof payload === "object" &&
                args.every(function (ref) {
                    var id = ref.id;
                    var test = ref.test;

                    return test(payload[id]);
                    }) &&
                Object.keys(payload).length === args.length; }
        }
        return obj
    }, {})

    var test = function (action) {
        var ref = unformat(action);
        var type = ref.type;
        var payload = ref.payload;
        return tests[type] && tests[type](payload)
    }

    return {
        schema: schemaMap,
        createMiddleware: middlewareHelper(tests, unformat),
        createReducer: reducerHelper(actions, unformat),
        actions: actions, test: test, actionCreators: actionCreators,
    }
}

exports.makeSchema = makeSchema;
exports.types = types;